/// @file        HwAdapt.c
/// @copyright     Copyright (c) 2012 Toradex AG \n
///             All information contained herein is, and remains the property of Toradex Ag and its suppliers, if any. The intellectual\n
///             and technical concepts contained herein are proprietary to Toradex AG and its suppliers and may be covered by Swiss and\n
///             foreign patents, patents in process, and are protected by trade secret or copyright law. Dissemination of this information\n
///             or reproduction of this material is strictly forbidden unless prior written permission is obtained from Toradex AG.\n
/// @author      $Author$
/// @version     $Rev$
/// @date        $Date$
/// @brief       The hardware adaption of the multitouch driver solution is made to run as an application.\n
///             The advantage is that a customer can modify, run and test the "driver" like a normal application. \n
///             The "hardware adaption" handles the whole interface to the touch controller (interrupt, I2C communication etc.)
///             using the standard Toradex library which are free available. Therefore it is much easier for a customer to adapt this
///             "hardware adaption" to other touch controller. \n
///                The counterpart to this "hardware adaption" is the "unified driver" which must be installed like a normal driver. \n
///                The "unified driver" has no connection to the hardware and run on all colibri modules and WinCE versions.
/// @target        Colibri T20, T30, WinEC7
/// @caveats    No screen output. All messages are send to the debug port.\n
///                This file contains the main loop.\n
///
#include <windows.h>
#include <Winbase.h>
#include <Winuser.h>
#include <winioctl.h>
#include <stdio.h>
#include "StdAfx.h"
#include "TchContr_EPxxxxM06.h"

//==== Debug Zones ================================================================================
// simple handling: debug zones are forced with TREU or FALSE and need a recompile to be activated
// for sophisticated handling: use macro DEBUGZONE(), DBGPARAM dpCurSettings etc.
// (see MSDN e.g. Win CE 5.0, Debug Zones)
#undef ZONE_ERROR
#undef ZONE_WARN
#undef ZONE_FUNCTION
#undef ZONE_INFO
#undef ZONE_TIPSTATE

#define ZONE_ERROR      TRUE
#define ZONE_WARN       FALSE
#define ZONE_FUNCTION   FALSE
#define ZONE_INFO       FALSE
#define ZONE_TIPSTATE   FALSE

//==== privat preprocessor definitions ============================================================
#define    COMTST_INT        5000                                             // timeout wait for comunication test (INFINITE = no timrout)

//==== public extern methods =====================================================================
// see "TchContr_EP0700.h"

//=================================================================================================
/// _tmain
int WINAPI WinMain( HINSTANCE hInstance,
                    HINSTANCE hPrevInstance,
                    LPTSTR    lpCmdLine,
                    int       nCmdShow)
{
    HANDLE            hDS;                                                // handle for "unified driver"
    CETCHINP_IOCTL    CeTchInp;                                            // touch events packed for IOCTL
    DWORD            BytesHandled;                                        // dummy for DeviceIoControl()
    DWORD            dwTimeout;                                            // timeout
    UINT            uiErrCnt;                                            // count of errors

    RETAILMSG(1,(TEXT("MultiTchHwAdapt loading...")));

    // Read the and setup the configuration
    vI2C_ReadConfiguration();                                           // read staic global configuration paramters (default or set in the regitry)

    // Init the IC interface on the Colibri (makes no sense to retry)
    if (FALSE==bI2C_Init())                                             // init I2C communication and setup interrupt handling
    {
        RETAILMSG(1,(TEXT("aborted\r\n")));                                // bI2C_Init() printed the detail of the failure alreday
        vI2C_DeInit();                                                    // release anythings activ allready
        return 0;
    }

    // Get a handle to the driver. Which causes the driver's XXX_Open function to be called
    hDS = CreateFile( TEXT("TCH1:"), GENERIC_READ|GENERIC_WRITE, FILE_SHARE_READ|FILE_SHARE_WRITE, NULL, OPEN_EXISTING, 0, 0);
    if (INVALID_HANDLE_VALUE==hDS)
    {
        RETAILMSG(1,(TEXT("no Unified Touch Driver...")));
        RETAILMSG(1,(TEXT("aborted\r\n")));                                // bI2C_Init() printed the detail of the failure alreday
        vI2C_DeInit();                                                    // release anythings activ allready
        return 0;
    }


    // main loop
    /// The used touch controller is setup to set the interrupt line activ if one or more fingers touch the screen.\n
    /// If the touch aera is untouched then the interrupt is set to passiv.
    /// The thread mainly sleeps until a timeout occurs. If this happen the interrupt signal is tested.
    /// If the interrupt line is activ (touches) then the status, current position etc. are read out\n
    /// of the touch controller and are processed. Afterwards the next timeout is set to the sample rate\n
    /// and the thread is going to sleep again.
    /// If the interrupt line is passiv (no touches) then the communcation to the touch controller is checked.\n
    /// Afterwards the next timeout is set to the communcation test rate.
    /// Beside that, a transition of the interrupt line generates a interrupt event which wakup the thread immideatly.\n

#ifdef TRACE_ON                                                            // Trace on
//    TraceCtrlClear();
    TraceTchClear();
//    TraceLastIdxClear();
#endif

    // main loop
    while (1==1)
    {
        //------ Startup Procedure --------------------------------------------------------------------

        DEBUGMSG(ZONE_INFO,(TEXT("MultiTchHwAdapt: Start communcation with the touch controller Fusion\r\n")));

        if (FALSE==bInitEPxxxxM06())                                    // reset and start touch controller
        {
            goto restart;
        }

        RETAILMSG(TRUE,(TEXT("done\r\n")));                                // app driver runs

        uiErrCnt = 5;                                                    // restart after 5 error

        dwTimeout = COMTST_INT;                                            // set timeout

        while(1==1)
        {
            if (TRUE==bWaitTouchDataINT(dwTimeout))                        // wit for interrupt with timeout
            {
                // interrupt and interruptsignal akctiv -> get the first touch events (normal state for the first touch)
                if (TRUE==bTouchPanelGetPoint(&CeTchInp.TchInp[0],&CeTchInp.iCnt)) // get the touch events from the controller
                {
                    // send the touch events to the unified touch driver
                    DeviceIoControl( hDS, IOCTL_SET_TOUCH_EVENT, &CeTchInp, sizeof(CeTchInp), NULL, 0, &BytesHandled, NULL );

#ifdef TRACE_ON                                                            // Trace on
//                    TraceCtrlPut();
                    TraceTchPut(&CeTchInp);
//                    TraceLastIdxPut();
#endif
                }
                else
                {
                    uiErrCnt--;
                    if (0>=uiErrCnt)                                    // count of error reached ?
                    {
                        goto restart;
                    }
                }
            }
            else                                                        // timeout
            {
                if (FALSE==bCommunicationTestEPxxxxM06())                // timeout, test the communication to the touch controller
                {
                    goto restart;
                }

#ifdef TRACE_ON                                                            // Trace on

//                TraceCtrlPrint();
//                TraceCtrlClear();
                TraceTchPrint();
                TraceTchClear();
//                TraceLastIdxPrint();
//                TraceLastIdxClear();
#endif
            }
        }


        // restart sequence
restart:
        RETAILMSG(1,(TEXT("MultiTchHwAdapt: restarts communication with touch controller...\r\n")));

        // send touch up event for currently all active touches
        vGenTouchUp(&CeTchInp.TchInp[0],&CeTchInp.iCnt);                // generate for all active touch a touch up event
        if (0<CeTchInp.iCnt)
        {
            // send the touch events to the unified touch driver
            DeviceIoControl( hDS, IOCTL_SET_TOUCH_EVENT, &CeTchInp, sizeof(CeTchInp), NULL, 0, &BytesHandled, NULL );
        }

        vDeInitEPxxxxM06();                                                // stop (reset/powerdown) the Touch Controller

        Sleep(400);                                                     // wait before an new try starts
    }
    return 0;
}
