/// @file        TchContr_Fusion.c
/// @copyright     Copyright (c) 2012 Toradex AG \n
///             All information contained herein is, and remains the property of Toradex Ag and its suppliers, if any. The intellectual\n
///             and technical concepts contained herein are proprietary to Toradex AG and its suppliers and may be covered by Swiss and\n
///             foreign patents, patents in process, and are protected by trade secret or copyright law. Dissemination of this information\n
///             or reproduction of this material is strictly forbidden unless prior written permission is obtained from Toradex AG.\n
/// @author      $Author$
/// @version     $Rev$
/// @date        $Date$
/// @brief       Fuction library for the touch controller Fusion
/// @target        Colibri PXA270, PX3xx, T20 WinCE5, WinCE6, WinEC7
/// @caveats    This file is used in the driver and in the driver utilities.\n
///             DRVUTIL markes the use in the utilities (mostly additional messages are printed.)\n
///             DRVUTIL is defined in the project properties ->C/C11++ ->Preprocessor Definition.

#include <windows.h>
#include <winioctl.h>
#include <winuser.h>

#include "StdAfx.h"
#include "TchContr_EPxxxxM06.h"
#include "IntLib.h"
#include "i2cLib.h"
#include "GpioLib.h"


//==== Debug Zones ================================================================================
// simple handling: debug zones are forced with TREU or FALSE and need a recompile to be activated
// for sophisticated handling: use macro DEBUGZONE(), DBGPARAM dpCurSettings etc.
// (see MSDN e.g. Win CE 5.0, Debug Zones)
#undef ZONE_ERROR
#undef ZONE_WARN
#undef ZONE_FUNCTION
#undef ZONE_INFO
#undef ZONE_TIPSTATE

#define ZONE_ERROR                    TRUE
#define ZONE_WARN                    TRUE
#define ZONE_FUNCTION                FALSE
#define ZONE_INFO                    FALSE
#define ZONE_TIPSTATE                FALSE


//==== privat preprocessor definitions ============================================================
// #define dwTRIGGER_PIN                106                                // define the pin for trigger

//-- I2C Communication ----------------------------------------------------------------------------
#define I2C_ADR                        0x38                                // I2C address of the Fusion chip

#define INT_PIN                        28                                    // default pin for the "INT Line"
#define INT_SIG_INV                    0                                    // 0 => INT signal is as described in the data sheet of the touch controller
                                                                            // >0 => INT signal is inverted
#define RESET_PIN                    30                                    // default pin for the "RESET Line"
#define RESET_POSTDLY                300                                    // delay after the "RESET Line" is released
#define RESET_LIN_INV                0                                    // 0 => "RESET Line" is as described in the data sheet of the touch controller
                                                                            // >0 => "RESET line" is inverted

#define I2C_100KB                    1                                    // 100kb I2C speed
#define I2C_400KB                    2                                    // 400kb I2C speed
#define HWADAPTPRIO                    THREAD_PRIORITY_ABOVE_NORMAL        // task priority hardware adapter

#define dwTHRESHOLD                    40                                    //
#define dwGAIN                       8                                    //
#define dwOFFSET                      0                                    //

#define ucREPRATE                    8                                    // repeat rate (is also used for communicaton test)


//-- "CTP Register Copy" --------------------------------------------------------------------------
// Touch sample block definition see EPxxxxM06.PDF / page 9
#define TCHSAMPL_SIZE                26                                    // size of touch sample block

#define HEAD_ADR                    0                                    // start header
#define PCKT_LENGTH_ADR                2                                    // start packet lenght
#define PNTCNT_ADR                    3                                    // start of the touch points
#define PNTCNT_SIZE                    1                                    // size of touch point data
#define CRC_ADR                        25                                    // start of CRC

#define ucTchPnt_Cnt_GET()            (0x0F&ucTouchSample[PNTCNT_ADR])    // get the count of touch points
#define ucTchPnt_Cnt_CLR()            (ucTouchSample[PNTCNT_ADR]=0)        // clr the count of touch points


#define TCHPNT_ADR                    5                                    // start of the touch points (relativ to the TCHSAMPL_ADR)
#define TCHPNT_SIZE                    4                                    // size of touch point data

#define uiPosXx_GET(fnr)            (((uint)ucTouchSample[TCHPNT_ADR+((fnr-1)*TCHPNT_SIZE)+1])+(256*(uint)(0x0F&ucTouchSample[TCHPNT_ADR+((fnr-1)*TCHPNT_SIZE)+0])))
#define uiPosYx_GET(fnr)            (((uint)ucTouchSample[TCHPNT_ADR+((fnr-1)*TCHPNT_SIZE)+3])+(256*(uint)(0x0F&ucTouchSample[TCHPNT_ADR+((fnr-1)*TCHPNT_SIZE)+2])))

#define    ucEvent_GET(fnr)            (0xC0&ucTouchSample[TCHPNT_ADR+((fnr-1)*TCHPNT_SIZE)+0])
#define TCH_DOWN                    0x00
#define TCH_UP                        0x40
#define TCH_MOVE                    0x80
#define RESERVED                    0xc0

#define    ucTouchID_GET(fnr)            (0xF0&ucTouchSample[TCHPNT_ADR+((fnr-1)*TCHPNT_SIZE)+2])

// addition register according sample code (no additional documentation !)
#define WORK_REGISTER_THRESHOLD        0x00
#define WORK_REGISTER_REPRATE        0x08
#define WORK_REGISTER_GAIN            0x30
#define WORK_REGISTER_OFFSET        0x31
#define WORK_REGISTER_NUM_X            0x33
#define WORK_REGISTER_NUM_Y            0x34

#define CMD_GET_FIRMWARE            0xbb
#define CMD_GET_SAMPLES                0xf9
#define CMD_READ_REGISTER            0xfc
#define CMD_READ_REGISTER_OFFSET    0x40

#define LASTCNT                        2                                    // retry count


//==== privat type definitions ====================================================================

//==== privat members =============================================================================
static DWORD                        dwI2C_Speed;
static DWORD                        dwI2C_Adr;

static DWORD                        dwINT_Pin;
static DWORD                        dwINT_Sig_Inv;
static DWORD                        dwRESET_Pin;
static DWORD                        dwRESET_PostDly;
static DWORD                        dwRESET_Lin_Inv;
static DWORD                        dwINT_GPIONumber;
static DWORD                        dwRESET_GPIONumber;

static HANDLE                       hI2C_Event;                            // for "interrupts"
static DWORD                        dwI2C_SysIntr;
static DWORD                        dwHwAdaptPrio;                        // task priority

static DWORD                        dwThreshold;
static DWORD                        dwGain;
static DWORD                        dwOffset;

static UCHAR                         ucTouchSample[TCHSAMPL_SIZE];        // copy of touch sample block
static UCHAR                         ucLastTouchSample[TCHSAMPL_SIZE];    // copy of touch sample block
static UINT8                        ucLastCnt;

static struct
{
    DWORD                            LastIndex;                            // register active touches
    DWORD                            x;                                    // last position
    DWORD                            y;
} LastTouches[MAXTOUCHES];

#ifdef dwTRIGGER_PIN
static DWORD                            dwTRIGGER_Pin;
static DWORD                            dwTRIGGER_GPIONumber;
#endif

//------ Trace Buffers ----------------------------------------------------------------------------
#ifdef TRACE_ON
static struct                                                            // trace for controller data
{
    UCHAR    TouchPoints;
    UINT    X1;
    UINT    Y1;
    UCHAR    Event1;
    UCHAR    TouchID1;
    UINT    X2;
    UINT    Y2;
    UCHAR    Event2;
    UCHAR    TouchID2;
    UINT    X3;
    UINT    Y3;
    UCHAR    Event3;
    UCHAR    TouchID3;
    UINT    X4;
    UINT    Y4;
    UCHAR    Event4;
    UCHAR    TouchID4;
}TraceCtrlBuf[200];
static UINT iCtrlIndex;

static struct                                                            // trace for last index
{
    CETCHINP_IOCTL                    TchInpCtrl;                            // touch events and counts
    DWORD                            dwTicks;
}TraceTchBuf[200];

static UINT iTchIndex;                                                    // index for trace buffer

static struct                                                            // trace for last index
{
    DWORD Index[MAXTOUCHES];
}TraceLastIdxBuf[200];
static UINT iLastIdxIndex;

#endif




//==== privat methods definitions ================================================================
// @internal
/// Read a register with one paramter
/// The index of this array is equal to the addresses of the "Work Mode Registers" in the touch controller.
/// @param[in]    sStartRegister    The device's internal register nummer
/// @param[in]    iNumber            Number of bytes to read
/// @param[out]   pucParam            Pointer to Parameter
/// @retval       TRUE  success
/// @retval       FALSE error, a message to the console shows the reason
/// @caveats      none
static BOOL bReadInReg(UCHAR ucRegNo,UCHAR *pucParam)
{
    LONG lErr;

    union                                                                // the Tegra I2C library needs an alingment of 4 for the buffer
    {                                                                    // therefore this union plac the char arry to an alignment address
        DWORD dwBuffer;
        UCHAR ucBuffer[3];
    }uAlingBuffer;

    if (FALSE==GetI2CLock(1000))
    {
        lErr = GetLastError();
        DEBUGMSG(ZONE_ERROR,(TEXT("MultiTchHwAdapt:bReadInReg() illegale response in GetI2Lock() %d\r\n"),lErr));
        Sleep(50);
        return FALSE;
    }

    uAlingBuffer.ucBuffer[0] = CMD_READ_REGISTER;                        // ??
    uAlingBuffer.ucBuffer[1] = CMD_READ_REGISTER_OFFSET+ucRegNo;

    if (FALSE == I2CBurstWrite((UCHAR)dwI2C_Adr,&uAlingBuffer.ucBuffer[0],-1,2))
    {
        lErr = GetLastError();
        RETAILMSG (TRUE,(TEXT("\r\nillegale response in I2CBurstWrite() %d"),lErr));
        ReleaseI2CLock();
        return FALSE;
    }


    if (FALSE == I2CBurstRead((UCHAR)dwI2C_Adr,&uAlingBuffer.ucBuffer[0],-1,1))
    {
        lErr = GetLastError();
        DEBUGMSG(ZONE_ERROR,(TEXT("MultiTchHwAdapt:bReadInReg() illegale response in I2CBurstRead() %d\r\n"),lErr));
        ReleaseI2CLock();
        return FALSE;
    }

    *pucParam = uAlingBuffer.ucBuffer[0];                                // copy the result

    ReleaseI2CLock();
    return TRUE;
}



// @internal
/// Write to a register with checksum
/// @param[in]    ucRegNo    The device's internal register nummer
/// @param[in]    ucParam    Parameter to send
/// @param[out]   none
/// @retval       TRUE  success
/// @retval       FALSE error, a message to the console shows the reason
/// @caveats      none
static BOOL bWriteOutReg(UCHAR ucRegNo,UCHAR ucParam)
{
    LONG lErr;

    union                                                                 // the Tegra I2C library needs an alingment of 4 for the buffer
    {                                                                    // therefore this union plac the char arry to an alignment address
        DWORD dwBuffer;
        UCHAR ucBuffer[4];
    } uAlingBuffer;

    if (FALSE==GetI2CLock(1000))
    {
        lErr = GetLastError();
        DEBUGMSG(ZONE_ERROR,(TEXT("MultiTchHwAdapt:bWriteOutReg() illegale response in GetI2Lock() %d\r\n"),lErr));
        Sleep(50);
        return FALSE;
    }

    uAlingBuffer.ucBuffer[0] = CMD_READ_REGISTER;                        // ??
    uAlingBuffer.ucBuffer[1] = ucRegNo;
    uAlingBuffer.ucBuffer[2] = ucParam;
    uAlingBuffer.ucBuffer[3] = uAlingBuffer.ucBuffer[0]^uAlingBuffer.ucBuffer[1]^uAlingBuffer.ucBuffer[2]; // checksumme

    if (FALSE == I2CBurstWrite((UCHAR)dwI2C_Adr,&uAlingBuffer.ucBuffer[0],-1,4))
    {
        lErr = GetLastError();
        DEBUGMSG(ZONE_ERROR,(TEXT("MultiTchHwAdapt:bWriteOutReg() illegale response in I2CBurstWrite() %d\r\n"),lErr));
        ReleaseI2CLock();
        return FALSE;
    }

    ReleaseI2CLock();
    return TRUE;
}


/// Read out all the registers containing touch data of the touch controller
/// @param[in]    none
/// @param[out]   if return TRUE then the global ucTouchSample[] contains the touch events read out of the controller
/// @retval       TRUE  success
/// @retval       FALSE error, the data fetch is failed. A message to the console shows the reason. ucTouchSample[] is undefined.
/// @caveats      global variables ucLastTouchSamples[] and ucLastCnt must be defined
static BOOL bGetTouchDataEPxxxxM06(void)
{
    LONG    lErr;
    UINT32    i;
    UINT8    crc;
    BOOL    rc;

    union                                                                // the Tegra I2C library needs an alingment of 4 for the buffer
    {                                                                    // therefore this union plac the char arry to an alignment address
        DWORD dwBuffer;
        UCHAR ucBuffer[TCHSAMPL_SIZE];
    }uAlingBuffer;

    rc = FALSE;                                                            // default is false


    if (FALSE==GetI2CLock(1000))
    {
        lErr = GetLastError();
        ERRORMSG(ZONE_ERROR, (TEXT("MultiTchHwAdapt:vGetControllerData() I2CBurstWrite() failed\r\n")));
        Sleep(50);
        goto quit;
    }

    // read samples
    uAlingBuffer.ucBuffer[0] = CMD_GET_SAMPLES;

    if (FALSE == I2CBurstWrite((UCHAR)dwI2C_Adr,&uAlingBuffer.ucBuffer[0],-1,1))
    {
        lErr = GetLastError();
        ERRORMSG(ZONE_ERROR, (TEXT("MultiTchHwAdapt:vGetControllerData() I2CBurstWrite() failed\r\n")));
        ReleaseI2CLock();
        goto quit;
    }

    memset(&uAlingBuffer.ucBuffer[0], 0, TCHSAMPL_SIZE);                // clear Buffer

    if (FALSE == I2CBurstRead((UCHAR)dwI2C_Adr,&uAlingBuffer.ucBuffer[0],-1,TCHSAMPL_SIZE))
    {
        lErr = GetLastError();
        ERRORMSG(ZONE_ERROR, (TEXT("MultiTchHwAdapt:vGetControllerData() I2CBurstRead() failed\r\n")));
        ReleaseI2CLock();
        goto quit;
    }

    ReleaseI2CLock();

    // test the Header and Packet lengt
    if ((0xAA!=uAlingBuffer.ucBuffer[HEAD_ADR]) || (0xAA!=uAlingBuffer.ucBuffer[HEAD_ADR+1]) || (0x1A!=uAlingBuffer.ucBuffer[PCKT_LENGTH_ADR]))
    {
        goto quit;
    }

    // calculate & check crc
    for (crc = 0, i = 0; i< CRC_ADR; i++)
    {
        crc^=uAlingBuffer.ucBuffer[i];
    }

    if (crc != uAlingBuffer.ucBuffer[CRC_ADR])
    {
        // CRC not valid
        // To avoid a touch up in case of a short I2C communication problem (glitches) up to 2 times the old values is talken
        if (ucLastCnt)
        {
            ERRORMSG(ZONE_ERROR, (TEXT("MultiTchHwAdapt:vGetControllerData() CRC fault use last values\r\n")));
            memcpy((void *)ucTouchSample,(void *)ucLastTouchSample,TCHSAMPL_SIZE); // use the last sample
            ucLastCnt--;
            rc = TRUE;                                                    // simulate data ok
        }
        else
        {
            ERRORMSG(ZONE_ERROR, (TEXT("MultiTchHwAdapt:vGetControllerData() CRC fault\r\n")));
            memset((void *)ucLastTouchSample, 0, TCHSAMPL_SIZE);        // clear the last samples for the next event
        }
    }
    else
    {
        // CRC ok
        memcpy((void *)ucTouchSample,(void *)&uAlingBuffer.ucBuffer[0],TCHSAMPL_SIZE); // copy the alignment buffer to touch sample block
        memcpy((void *)ucLastTouchSample, (void *)&uAlingBuffer.ucBuffer[0], TCHSAMPL_SIZE); // save it in last sample
        ucLastCnt = LASTCNT;                                            // reset retry counter
        rc = TRUE;                                                        // data read ok
    }

quit:
//    InterruptDoneCompat(dwI2C_SysIntr);                                    // interrupt quit

    return rc;
}


// @internal
/// Read the firmware version
/// @param[in]    none
/// @param[out]   pucFirmwareStr String pointer to store the frirmware string (22 byte at least)
/// @retval       TRUE  success
/// @retval       FALSE error, a message to the console shows the reason
/// @caveats      none
BOOL bReadFirmwareEPxxxxM06(UCHAR *pucFirmwareStr)
{
    LONG lErr;

    union                                                                // the Tegra I2C library needs an alingment of 4 for the buffer
    {                                                                    // therefore this union plac the char arry to an alignment address
        DWORD dwBuffer;
        UCHAR ucBuffer[22+1];                                            // add one char for end string (for security reasion)
    }uAlingBuffer;

    if (FALSE==GetI2CLock(1000))
    {
        lErr = GetLastError();
        RETAILMSG (TRUE,(TEXT("\r\nillegale response in GetI2Lock() %d"),lErr));
        Sleep(50);
        return FALSE;
    }

    uAlingBuffer.ucBuffer[0] = CMD_GET_FIRMWARE;

    if (FALSE == I2CBurstWrite((UCHAR)dwI2C_Adr,&uAlingBuffer.ucBuffer[0],-1,1))
    {
        lErr = GetLastError();
        RETAILMSG (TRUE,(TEXT("\r\nillegale response in I2CBurstWrite() %d"),lErr));
        ReleaseI2CLock();
        return FALSE;
    }

    memset(&uAlingBuffer.ucBuffer[0], 0, sizeof(uAlingBuffer.ucBuffer));

    if (FALSE == I2CBurstRead((UCHAR)dwI2C_Adr,&uAlingBuffer.ucBuffer[0],-1,22))
    {
        lErr = GetLastError();
        RETAILMSG (TRUE,(TEXT("\r\nillegale response in I2CBurstRead() %d"),lErr));
        ReleaseI2CLock();
        return FALSE;

    }

    memcpy((void *)pucFirmwareStr,(void *)&uAlingBuffer.ucBuffer[0],22);    // copy the alignment buffer to orginal place

    ReleaseI2CLock();
    return TRUE;
}




#ifdef dwTRIGGER_PIN
/// Set the trigger output to high

/// Set the trigger output to high
/// @param[in]    none
/// @param[out]   none
/// @retval       none
/// @caveats      none
void vSetTriggerHigh(void)
{
    SetGPIOLevel (dwTRIGGER_GPIONumber,1);
}

/// Set the trigger output to low
/// @param[in]    none
/// @param[out]   none
/// @retval       none
/// @caveats      none
void vSetTriggerLow(void)
{
    SetGPIOLevel (dwTRIGGER_GPIONumber,0);
}
#endif



//==== public preprocessor definitions ===========================================================
//==== public members ============================================================================
//==== public methods definitions ================================================================


/// Set the configuration data according to the registry entry or to the default value
void vI2C_ReadConfiguration(void)
{
    HKEY    hKey;
    DWORD   dwSize;
    LONG    lError;

    // set default registry values
    dwI2C_Speed            = I2C_400KB;                                    // I2C speed
    dwI2C_Adr           = I2C_ADR;                                        // I2C address of the TangoC48 chip

    dwINT_Pin            = INT_PIN;                                        // pin for Int
    dwINT_Sig_Inv        = INT_SIG_INV;                                    // Int signal invertetd
    dwRESET_Pin            = RESET_PIN;                                    // pin for Reset
    dwRESET_PostDly        = RESET_POSTDLY;                                // delay after the "RESET Line" is released
    dwRESET_Lin_Inv        = RESET_LIN_INV;                                // Reset line inverted
    dwHwAdaptPrio        = HWADAPTPRIO;                                    // Task priority
    dwINT_GPIONumber    = 0;                                            // GPIO pin for Int
    dwRESET_GPIONumber  = 0;                                            // GPIO pin for Reset

    dwThreshold            = dwTHRESHOLD;                                    // see PolyTouch Application Note, chapter 5.1
    dwGain                = dwGAIN;                                        // see PolyTouch Application Note, chapter 5.1
    dwOffset            = dwOFFSET;                                        // see PolyTouch Application Note, chapter 5.1


#ifdef dwTRIGGER_PIN
    dwTRIGGER_Pin    = dwTRIGGER_PIN;
#endif

    // lock if there are settings in the registry
    lError = RegOpenKeyEx(HKEY_LOCAL_MACHINE, L"Software\\Toradex\\MultiTchHwAdaptEPxxxxM06", 0, KEY_ALL_ACCESS, &hKey);
    if (lError == ERROR_SUCCESS)                                        // only if the registry key exitsts
    {
        dwSize=sizeof(DWORD);
        RegQueryValueEx(hKey, L"Int_Pin", NULL, NULL, (LPBYTE)&dwINT_Pin, &dwSize); // get the GPIONumber for interrupt
        RegQueryValueEx(hKey, L"Int_Signal_Inv", NULL, NULL, (LPBYTE)&dwINT_Sig_Inv, &dwSize); // get the GPIONumber for interrupt active level
        RegQueryValueEx(hKey, L"Reset_Pin", NULL, NULL, (LPBYTE)&dwRESET_Pin, &dwSize); // get the GPIONumber for reset
        RegQueryValueEx(hKey, L"Reset_Post_Delay", NULL, NULL, (LPBYTE)&dwRESET_PostDly, &dwSize); // get the pso delay for reset
        RegQueryValueEx(hKey, L"Reset_Line_Inv", NULL, NULL, (LPBYTE)&dwRESET_Lin_Inv, &dwSize); // get the inversion for reset
        RegQueryValueEx(hKey, L"I2CSpeed", NULL, NULL, (LPBYTE)&dwI2C_Speed,&dwSize); // get the I2C Speed
        RegQueryValueEx(hKey, L"I2CAddress", NULL, NULL, (LPBYTE)&dwI2C_Adr,&dwSize); // get the I2C Address
        RegQueryValueEx(hKey, L"HwAdaptPriority", NULL, NULL, (LPBYTE)&dwHwAdaptPrio,&dwSize); // get the priority

        RegQueryValueEx(hKey, L"Threshold", NULL, NULL, (LPBYTE)&dwThreshold,&dwSize); // get the Threshold
        RegQueryValueEx(hKey, L"Gain", NULL, NULL, (LPBYTE)&dwGain,&dwSize); // get the Gain
        RegQueryValueEx(hKey, L"Offset", NULL, NULL, (LPBYTE)&dwOffset,&dwSize); // get the Offset
        RegCloseKey (hKey);
    }
}

/// Init the I2C communcation incl.the GIPO for the "INT" and "RST" pin of the touch controller
BOOL bI2C_Init(void) //ok//
{
    BOOL        bSuccess;
    DWORD       dwVerMaj;
    DWORD       dwVerMin;
    DWORD       dwVerBuild;
    DWORD       dwIRq;
    DWORD       dwEdge;

    PIN_INSTANCE sGPIO;

    //------ I2C init ----------------------------------------------------------------------------
    I2cIntGetLibVersion(&dwVerMaj, &dwVerMin, &dwVerBuild);

    DEBUGMSG(ZONE_INFO,(TEXT("MultiTchAppDriv:I2C Library Version=%d.%d Build=%d\r\n"),dwVerMaj, dwVerMin, dwVerBuild));

    bSuccess = FALSE;

    if (TRUE==I2CInit())
    {
        switch (dwI2C_Speed)
        {
            case I2C_100KB:
                bSuccess = GetI2CLock(500);
                if (TRUE==bSuccess)
                {
                    I2CSetSpeed(I2C_100KBPS);                            // set speed
                    ReleaseI2CLock();
                }
                break;

            case I2C_400KB:
                bSuccess = GetI2CLock(500);
                if (TRUE==bSuccess)
                {
                    I2CSetSpeed(I2C_400KBPS);                            // set speed
                    ReleaseI2CLock();
                }
                break;

            default:
                break;
        }
    }

    //------ GPIO init ----------------------------------------------------------------------------
    if (TRUE==bSuccess)
    {
        bSuccess = FALSE;

        if (TRUE == GetGPIOFromPin(dwINT_Pin, FALSE, &sGPIO))
        {
            if (0xFF != sGPIO.inst2)                                    // function on a second GPIO exist
            {
                SetGPIOAltFn((DWORD)sGPIO.inst2, -1, DIR_IN);            // set it to input
            }
            dwINT_GPIONumber = sGPIO.inst1;                                // use inst1 as GPIONumber

            if (TRUE == GetGPIOFromPin(dwRESET_Pin, FALSE, &sGPIO))
            {
                if (0xFF != sGPIO.inst2)                                // function on a second GPIO exist
                {
                    SetGPIOAltFn((DWORD)sGPIO.inst2, -1, DIR_IN);        // set it to input
                }
                dwRESET_GPIONumber = sGPIO.inst1;                        // use inst1 as GPIONumber
                bSuccess = TRUE;
            }
            else
            {
                DEBUGMSG(ZONE_ERROR,(TEXT("Multi-Touch HW-Adaption_Init:bI2C_Init() look up Reset_GPIONumber fails\r\n")));
            }
        }
        else
        {
            DEBUGMSG(ZONE_ERROR,(TEXT("Multi-Touch HW-Adaption_Init:bI2C_Init() look up Int_GPIONumber fails\r\n")));
        }
    }

    if (TRUE==bSuccess)
    {
        bSuccess = FALSE;

        SetGPIOAltFn(dwRESET_GPIONumber, -1, DIR_OUT);
        if (0==dwRESET_Lin_Inv)                                            // Reset not inverted
        {
            SetGPIOLevel (dwRESET_GPIONumber,0);                        // Reset activ / no power (vInitTangoC48())
        }
        else
        {
            SetGPIOLevel (dwRESET_GPIONumber,1);                        // Reset activ / no power (vInitTangoC48())
        }

        SetGPIOAltFn(dwINT_GPIONumber, -1, DIR_IN);                        // Set INT as Input

        dwIRq=GetGPIOIrq(dwINT_GPIONumber);
        if(dwIRq)
        {
            // Set edge to trigger (Rising, Falling or both)
            // dwEdge=GPIO_EDGE_RISING | GPIO_EDGE_FALLING;                // (rising & falling edge)
            dwEdge=GPIO_EDGE_FALLING;

            // dwEdge=GPIO_EDGE_RISING;

            if(SetGPIOIrqEdge(dwINT_GPIONumber,dwEdge))
            {
                // Create an event to wait on
                hI2C_Event = CreateEvent(NULL, FALSE, FALSE, NULL);
                if (hI2C_Event!= NULL)
                {
                    // Get the SYSINTR that corresponds to dwIrq (non GPIO Irqs corresond to the bit numbers of the PXA27x Interrupt resisters)
                    dwI2C_SysIntr=RequestSysInterrupt(dwIRq);

                    if (dwI2C_SysIntr)
                    {
                        if (InterruptInitializeCompat(dwI2C_SysIntr,hI2C_Event,NULL,0))
                        {
                            // set interrupt priority
                            CeSetThreadPriority(GetCurrentThread(), dwHwAdaptPrio);
                            bSuccess = TRUE;
                        }
                    }
                }
            }
        }
    }

    if (FALSE==bSuccess)
    {
        DEBUGMSG(ZONE_ERROR,(TEXT("Multi-Touch HW-Adaption_Init:bI2C_Init() had failed\r\n")));
    }

#ifdef dwTRIGGER_PIN
     if (TRUE==bSuccess)
    {
        // for Trigger Output
        if (TRUE == GetGPIOFromPin(dwTRIGGER_Pin, FALSE, &sGPIO))
        {
            if (0xFF != sGPIO.inst2)                                // function on a second GPIO exist
            {
                SetGPIOAltFn((DWORD)sGPIO.inst2, -1, DIR_IN);        // set it to input
            }
            dwTRIGGER_GPIONumber = sGPIO.inst1;                        // use inst1 as GPIONumber
            bSuccess = TRUE;
        }
        else
        {
            DEBUGMSG(ZONE_ERROR,(TEXT("MultiTchAppDriv:bI2C_Init() look up TRIGGER_GPIONumber fails\r\n")));
        }
        SetGPIOAltFn(dwTRIGGER_GPIONumber, -1, DIR_OUT);
        SetGPIOLevel (dwTRIGGER_GPIONumber,1);
    }
#endif

    return bSuccess;
}


/// Start the touch controller (release the RST/PWR line) start the commincation and and setup the
BOOL bInitEPxxxxM06(void)
{
    INT j;

    if (0==dwRESET_Lin_Inv)                                                // Reset not inverted
    {
        SetGPIOLevel (dwRESET_GPIONumber,1);                            // Reset passiv / Power on)
    }
    else
    {
        SetGPIOLevel (dwRESET_GPIONumber,0);                            // Reset passiv / Power on)
    }

    Sleep(dwRESET_PostDly);                                                // wait until the reset sequence of the touch controller is over

    //-------- touch controller --------------------------------------------------------------------
    if (FALSE==bWriteOutReg(WORK_REGISTER_GAIN,(UCHAR)dwGain))            // values according the driver exampel
    {
        return FALSE;
    }

    if (FALSE==bWriteOutReg(WORK_REGISTER_THRESHOLD,(UCHAR)dwThreshold))
    {
        return FALSE;
    }

    if (FALSE==bWriteOutReg(WORK_REGISTER_OFFSET,(UCHAR)dwOffset))
    {
        return FALSE;
    }

    if (FALSE==bWriteOutReg(WORK_REGISTER_REPRATE, ucREPRATE))
    {
        return FALSE;
    }

    // Init Finger data
    ucTchPnt_Cnt_CLR();                                                    // no touches cause a touch up event
    memset((void *)ucLastTouchSample, 0, TCHSAMPL_SIZE);                // clear the last samples for the next event
    ucLastCnt = LASTCNT;                                                // reset retry counter

    // reset last touches
    for(j=0;j<MAXTOUCHES;j++)
    {
        LastTouches[j].LastIndex    = -1;
        LastTouches[j].x            = 0;
        LastTouches[j].y            = 0;
    }


#if 0
    // get more informatuion like Firmware etc
#endif

    return TRUE;
}

/// Test the level of the INT line
BOOL bTstINTEPxxxxM06(void)
{
    if (0==dwINT_Sig_Inv)                                            // interrupt input is not inverted
    {
        return (0==GetGPIOLevel(dwINT_GPIONumber));                    // interrupt input allready high means there are pending touchs
    }
    else
    {
        return (1==GetGPIOLevel(dwINT_GPIONumber));                    // interrupt input allready low means there are pending touchs
    }
}


/// Release the current interrupt and wait for the next interrupt with timeout
BOOL bWaitTouchDataINT(DWORD dwTimeout)
{
    BOOL bRet;
    if (WAIT_TIMEOUT == WaitForSingleObject(hI2C_Event,dwTimeout))    // wait for input interrupt or timeout
    {
        bRet = FALSE;
    }
    else
    {
        bRet = TRUE;
    }
    InterruptDoneCompat(dwI2C_SysIntr);                                    // interrupt quit

    return bRet;
}

/// Test the communication by reading out the vedors's chp id
BOOL bCommunicationTestEPxxxxM06(void) //ok//
{
    UCHAR ucTemp;

    if (TRUE==bTstINTEPxxxxM06())                                        // Interrupt activ in idle ?
    {
        return FALSE;
    }
    else
    {
        // read the register "Report Rate" to be shure the controller gives an answer
        if (FALSE==bReadInReg(WORK_REGISTER_REPRATE, &ucTemp))
        {
            return FALSE;
        }
        else
        {
            if (ucREPRATE!=ucTemp)
            {
                return FALSE;
            }
            else
            {
                return TRUE;
            }
        }
    }
}

/// Get the touch events from the controller
BOOL bTouchPanelGetPoint(PTCHINPUT pInput,INT *pCnt)
{
    INT    i,j;
    INT rc;

    UCHAR ucTemp;

    PTCHINPUT pInput_tmp;                                                // for loops

    DEBUGMSG(ZONE_FUNCTION||ZONE_INFO, (TEXT("MultiTchAppDriv:vTouchPanelGetPoint() +\r\n")));

    // clear all data for the call back to the MDD and set count of touches to  0
    rc = TRUE;                                                            // default tue
    memset(pInput, 0, MAXTOUCHES*sizeof(TCHINPUT));                    // clear array for call back function
    *pCnt = 0;

    if (FALSE==bGetTouchDataEPxxxxM06())                                // get the data from the touch Controller
    {
        return FALSE;
    }

    // clear events
    pInput_tmp = pInput;                                                // copy pointer for loop

    for (i = 1; i<=MAXTOUCHES; i++)
    {
        //------ no more touches ------------------------------------------------------------------
        ucTemp = ucEvent_GET(i);

        if (RESERVED == ucEvent_GET(i))                                    // no more touches
        {
            break;
        }
        else
        {
            //------ touch up ---------------------------------------------------------------------
            if (TCH_UP == ucEvent_GET(i))                                // touch up
            {
                pInput_tmp->dwID    = (DWORD)ucTouchID_GET(i)+1;        // get the id
                pInput_tmp->dwFlags    = (DWORD)TOUCHEVENTF_UP;            // set the flags

                // ID of active touches are registered and cleared in case of touch up
                for (j = 0; j <MAXTOUCHES; j++)
                {
                    if (pInput_tmp->dwID == LastTouches[j].LastIndex)
                    {
                        // get the last position
                        pInput_tmp->x                = LastTouches[j].x;
                        pInput_tmp->y                = LastTouches[j].y;
                        // reset ID
                        LastTouches[j].x            = 0;
                        LastTouches[j].y            = 0;
                        LastTouches[j].LastIndex    = -1;
                        break;
                    }
                }
            }
            else
            {
                //------ touch down or move -------------------------------------------------------
                pInput_tmp->x        = (DWORD)uiPosXx_GET(i);            // get the position
                pInput_tmp->y        = (DWORD)uiPosYx_GET(i);
                pInput_tmp->dwID    = (DWORD)ucTouchID_GET(i)+1;        // get the id

                // the ID of active touches msut be registered
                for (j = 0; j <MAXTOUCHES; j++)                            // first find an "old" entry to register the touch
                {
                    if (pInput_tmp->dwID == LastTouches[j].LastIndex)
                    {
                        // save position. Index is saved already
                        LastTouches[j].x    = pInput_tmp->x;
                        LastTouches[j].y    = pInput_tmp->y;
                        break;
                    }
                }
                if (MAXTOUCHES<=j)                                        // no entries found the search for an empty place
                {
                    // find the next free entry to register the touch again
                    for (j = 0; j <MAXTOUCHES; j++)
                    {
                        if (-1 == LastTouches[j].LastIndex)
                        {
                            // save position and index
                            LastTouches[j].x            = pInput_tmp->x;
                            LastTouches[j].y            = pInput_tmp->y;
                            LastTouches[j].LastIndex    = pInput_tmp->dwID;
                            break;
                        }
                    }
                }
                if (MAXTOUCHES<=j)                                        // no empty place use the last one
                {
                    // save position and index
                    LastTouches[MAXTOUCHES-1].x            = pInput_tmp->x;
                    LastTouches[MAXTOUCHES-1].y            = pInput_tmp->y;
                    LastTouches[MAXTOUCHES-1].LastIndex    = pInput_tmp->dwID;
                }

                //------ touch down ---------------------------------------------------------------
                if (TCH_DOWN == ucEvent_GET(i))
                {
                    pInput_tmp->dwFlags    = (DWORD)(TOUCHEVENTF_INRANGE | TOUCHEVENTF_DOWN); // set the flags
                }
                else
                {
                    //------ touch move -----------------------------------------------------------
                    if (TCH_MOVE == ucEvent_GET(i))
                    {
                        pInput_tmp->dwFlags    = (DWORD)(TOUCHEVENTF_INRANGE | TOUCHEVENTF_MOVE); // set the flags
                    }
                    else
                    {
                        rc = FALSE;
                    }
                }
            }
        }

        pInput_tmp++;                                                    // next touch point
    }

    *pCnt = i-1;                                                        // correct and set the count of touches


#if 0
    else
    {
        DEBUGMSG(ZONE_INFO,(TEXT("MultiTchAppDriv:vTouchPanelGetPoint() invalid count of finges (%d)\r\n" ),iFiCnt));
    }
#endif

    DEBUGMSG(ZONE_FUNCTION||ZONE_INFO, (TEXT("MultiTchAppDriv:vTouchPanelGetPoint() -\r\n")));
    return rc;
}

///    Send touch up event for currently all active touches
void vGenTouchUp(PTCHINPUT pInput,INT *pCnt)
{
    INT    j;

    PTCHINPUT pInput_tmp;                                                // for loops

    DEBUGMSG(ZONE_FUNCTION||ZONE_INFO, (TEXT("MultiTchAppDriv:vTouchPanelGetPoint() +\r\n")));

    // clear all data for the call back to the MDD and set count of touches to  0
    memset(pInput, 0, MAXTOUCHES*sizeof(TCHINPUT));                        // clear array for call back function
    *pCnt = 0;

    // search for active touches
    pInput_tmp = pInput;                                                // copy pointer for loop

    for (j = 0; j<MAXTOUCHES; j++)
    {
        if (-1 !=LastTouches[j].LastIndex)                                // active touch ?
        {
            pInput_tmp->x        = LastTouches[j].x;                        // get the last position
            pInput_tmp->y        = LastTouches[j].y;
            pInput_tmp->dwID    = LastTouches[j].LastIndex;                // get the index
            pInput_tmp->dwFlags    = (DWORD)TOUCHEVENTF_UP;                // set the flags to touch up
            *pCnt++;
        }
        pInput_tmp++;
    }
}



/// Reset on or power off the touch controller
void vDeInitEPxxxxM06(void)                                                // stop (reset/powerdown) the Touch Controller
{
    if (0==dwRESET_Lin_Inv)                                                // Reset not inverted
    {
        SetGPIOLevel (dwRESET_GPIONumber,0);                            // Reset activ or no power
    }
    else
    {
        SetGPIOLevel (dwRESET_GPIONumber,1);                            // Reset activ or no power
    }

    ReleaseI2CLock();
}


/// Stop the I2C communication (release interrrupt, mutex, stop the I2C driver etc)
void vI2C_DeInit(void)
{
    // GPIO disabel
    InterruptDoneCompat(dwI2C_SysIntr);
    InterruptDisableCompat(dwI2C_SysIntr);
    ReleaseSysIntr(dwI2C_SysIntr);

    // I2C driver
    ReleaseI2CLock();
    I2CDeInit();
}


//------ Trace Functions --------------------------------------------------------------------------
#ifdef TRACE_ON

// reset index and whole controller buffer
void TraceCtrlClear(void)
{
    // erase trace buffer
    iCtrlIndex = 1000;
    memset(&TraceCtrlBuf[0], 0x55, sizeof(TraceCtrlBuf));
}


// reset index and whole touch buffer
void TraceTchClear(void)
{

    // erase trace buffer
    iTchIndex = 1000;
    memset(&TraceTchBuf[0], 0, sizeof(TraceTchBuf));
}

// reset index and whole last touch buffer
void TraceLastIdxClear(void)
{
    // erase trace buffer
    iLastIdxIndex = 1000;
    memset(&TraceLastIdxBuf[0], 0, sizeof(TraceLastIdxBuf));
}

// write controler data to the controller buffer
void TraceCtrlPut(void)
{
    if (1000==iCtrlIndex)
        iCtrlIndex=0;

    if (iCtrlIndex < 200)
    {
        TraceCtrlBuf[iCtrlIndex].TouchPoints    = ucTchPnt_Cnt_GET();
        TraceCtrlBuf[iCtrlIndex].X1                = uiPosXx_GET(1);
        TraceCtrlBuf[iCtrlIndex].Y1                = uiPosXx_GET(1);
        TraceCtrlBuf[iCtrlIndex].Event1            = ucEvent_GET(1);
        TraceCtrlBuf[iCtrlIndex].TouchID1        = ucTouchID_GET(1);
        TraceCtrlBuf[iCtrlIndex].X2                = uiPosXx_GET(2);
        TraceCtrlBuf[iCtrlIndex].Y2                = uiPosXx_GET(2);
        TraceCtrlBuf[iCtrlIndex].Event2            = ucEvent_GET(2);
        TraceCtrlBuf[iCtrlIndex].TouchID2        = ucTouchID_GET(2);
        TraceCtrlBuf[iCtrlIndex].X3                = uiPosXx_GET(3);
        TraceCtrlBuf[iCtrlIndex].Y3                = uiPosXx_GET(3);
        TraceCtrlBuf[iCtrlIndex].Event3            = ucEvent_GET(3);
        TraceCtrlBuf[iCtrlIndex].TouchID3        = ucTouchID_GET(3);
        TraceCtrlBuf[iCtrlIndex].X4                = uiPosXx_GET(4);
        TraceCtrlBuf[iCtrlIndex].Y4                = uiPosXx_GET(4);
        TraceCtrlBuf[iCtrlIndex].Event4            = ucEvent_GET(4);
        TraceCtrlBuf[iCtrlIndex].TouchID4        = ucTouchID_GET(4);
        iCtrlIndex++;
    }
}

// write touch data to the controller buffer
void TraceTchPut(CETCHINP_IOCTL *pTchBuf)
{
    if (1000 == iTchIndex)
        iTchIndex=0;

    if (iTchIndex < 200)
    {
        TraceTchBuf[iTchIndex].TchInpCtrl    = *pTchBuf;
        TraceTchBuf[iTchIndex].dwTicks        = GetTickCount();
        iTchIndex++;
    }
}

// write current touch to the last touch buffer
void TraceLastIdxPut(void)
{
    if (1000 == iLastIdxIndex)
        iLastIdxIndex=0;

    if (iLastIdxIndex < 200)
    {
        TraceLastIdxBuf[iLastIdxIndex].Index[0] = LastTouches[0].LastIndex;
        TraceLastIdxBuf[iLastIdxIndex].Index[1] = LastTouches[1].LastIndex;
        TraceLastIdxBuf[iLastIdxIndex].Index[2] = LastTouches[2].LastIndex;
        TraceLastIdxBuf[iLastIdxIndex].Index[3] = LastTouches[3].LastIndex;
        TraceLastIdxBuf[iLastIdxIndex].Index[4] = LastTouches[4].LastIndex;
        iLastIdxIndex++;
    }
}

// print out the whole controller buffer
void TraceCtrlPrint(void)
{
    UINT i;

    if (1000==iCtrlIndex)
        return;

    for (i = 0; i <iCtrlIndex; i++)
    {
        DEBUGMSG(ZONE_WARN,(TEXT("TraceCtrlIndex=%d  Touch Points=%d\r\n" ),i,TraceCtrlBuf[i].TouchPoints));
        DEBUGMSG(ZONE_WARN,(TEXT("TraceCtrlIndex=%d  TouchID1=%d, Event1=%d, X1=%d, Y1=%d \r\n" ),i,TraceCtrlBuf[i].TouchID1,TraceCtrlBuf[i].Event1,TraceCtrlBuf[i].X1,TraceCtrlBuf[i].Y1));
        DEBUGMSG(ZONE_WARN,(TEXT("TraceCtrlIndex=%d  TouchID2=%d, Event2=%d, X2=%d, Y2=%d \r\n" ),i,TraceCtrlBuf[i].TouchID2,TraceCtrlBuf[i].Event2,TraceCtrlBuf[i].X2,TraceCtrlBuf[i].Y2));
        DEBUGMSG(ZONE_WARN,(TEXT("TraceCtrlIndex=%d  TouchID3=%d, Event3=%d, X3=%d, Y3=%d \r\n" ),i,TraceCtrlBuf[i].TouchID3,TraceCtrlBuf[i].Event3,TraceCtrlBuf[i].X3,TraceCtrlBuf[i].Y3));
        DEBUGMSG(ZONE_WARN,(TEXT("TraceCtrlIndex=%d  TouchID4=%d, Event4=%d, X4=%d, Y3=%d \r\n" ),i,TraceCtrlBuf[i].TouchID4,TraceCtrlBuf[i].Event4,TraceCtrlBuf[i].X4,TraceCtrlBuf[i].Y4));
        DEBUGMSG(ZONE_WARN,(TEXT("\r\n" )));
        Sleep(10);
    }
    DEBUGMSG(ZONE_WARN,(TEXT("\r\n" )));
}

// print out the whole touch data buffer
void TraceTchPrint(void)
{
    UINT i,j,cnt;

    if (1000 == iTchIndex)
        return;

    for (i = 0; i <iTchIndex; i++)
    {
        cnt = TraceTchBuf[i].TchInpCtrl.iCnt;

        DEBUGMSG(ZONE_WARN,(TEXT("TraceTchIndex=%d, Touch Count=%d, Time=%d\r\n" ),i,cnt,TraceTchBuf[i].dwTicks));
        for (j = 0; j <cnt; j++)
        {
            DEBUGMSG(ZONE_WARN,(TEXT("TraceTchIndex=%d, ID=%d, Flags=%d, x=%d, y=%d\r\n" ),i,TraceTchBuf[i].TchInpCtrl.TchInp[j].dwID,TraceTchBuf[i].TchInpCtrl.TchInp[j].dwFlags,TraceTchBuf[i].TchInpCtrl.TchInp[j].x,TraceTchBuf[i].TchInpCtrl.TchInp[j].y));
            Sleep(10);
        }
    }
    DEBUGMSG(ZONE_WARN,(TEXT("\r\n" )));
}


// print out the current last touch buffer
void TraceLastIdxPrint(void)
{
    UINT i;

    if (1000 == iLastIdxIndex)
        return;

    for (i = 0; i <iLastIdxIndex; i++)
    {
        DEBUGMSG(ZONE_WARN,(TEXT("LastIndex=%d, 0=%d, 1=%d, 2=%d, 3=%d, 4=%d\r\n" ),i,TraceLastIdxBuf[i].Index[0],TraceLastIdxBuf[i].Index[1],TraceLastIdxBuf[i].Index[2],TraceLastIdxBuf[i].Index[3],TraceLastIdxBuf[i].Index[4]));
    }
    DEBUGMSG(ZONE_WARN,(TEXT("\r\n" )));
}
#endif




